from isomedia.atom import Atom, FullAtom, LazyLoadAtom, interpret_atom, write_atom

ISOM_ATOMS = [
    'bxml',
    'co64',
    'cprt',
    'cslg',
    'ctts',
    'dinf',
    'dref',
    'edts',
    'elst',
    'fecr',
    'fiin',
    'fire',
    'fpar',
    'free',
    'frma',
    'ftyp',
    'gitn',
    'hdlr',
    'hmhd',
    'idat',
    'iinf',
    'iloc',
    'ipro',
    'iref',
    'leva',
    'mdat',
    'mdhd',
    'mdia',
    'meco',
    'mehd',
    'mere',
    'meta',
    'mfhd',
    'mfra',
    'mfro',
    'minf',
    'moof',
    'moov',
    'mvex',
    'mvhd',
    'nmhd',
    'padb',
    'paen',
    'pdin',
    'pitm',
    'prft',
    'saio',
    'saiz',
    'sbgp',
    'schi',
    'schm',
    'sdtp',
    'segr',
    'sgpd',
    'sidx',
    'sinf',
    'skip',
    'smhd',
    'ssix',
    'stbl',
    'stco',
    'stdp',
    'strd',
    'stri',
    'strk',
    'stsc',
    'stsd',
    'stsh',
    'stss',
    'stsz',
    'stts',
    'styp',
    'stz2',
    'subs',
    'tfdt',
    'tfhd',
    'tfra',
    'tkhd',
    'traf',
    'trak',
    'tref',
    'trex',
    'trgr',
    'trun',
    'tsel',
    'udta',
    'vmhd',
    'xml ',
]

class FreeAtom(LazyLoadAtom):
    pass

class SkipAtom(LazyLoadAtom):
    pass

class MdatAtom(LazyLoadAtom):
    pass

class UserExtendedAtom(Atom):
    # TODO: How should I surface both uuid and the extended types
    def __init__(self, atom_header, atom_body, document, parent_atom, file_offset):
        Atom.__init__(self, atom_header, atom_body, document, parent_atom, file_offset)

        self._user_type = atom_body[0:16]
        self._data = atom_body

    def get_data(self):
        return self._data

    def to_bytes(self):
        written = Atom.to_bytes(self)
        return ''.join([written, self.get_data()])

class FtypAtom(Atom):
    def __init__(self, atom_header, atom_body, document, parent_atom, file_offset):
        Atom.__init__(self, atom_header, atom_body, document, parent_atom, file_offset)

        definition = [
            ('major_brand', (4, None)),
            ('minor_version', (4, int)),
            ('compatible_brands', (list, (4, None), None))
        ]

        self.properties.update(interpret_atom(atom_body, definition))
        self._definition['FtypAtom'] = definition

    def to_bytes(self):
        written = Atom.to_bytes(self)
        return ''.join([written, write_atom(self.properties, self._definition['FtypAtom'])])

class MvhdAtom(FullAtom):
    def __init__(self, atom_header, atom_body, document, parent_atom, file_offset):
        FullAtom.__init__(self, atom_header, atom_body, document, parent_atom, file_offset)

        definition = []

        if self.properties['version'] == 1:
            definition.extend([
                ('creation_time', (8, int)),
                ('modification_time', (8, int)),
                ('timescale', (4, int)),
                ('duration', (8, int))
            ])
        else:
            definition.extend([
                ('creation_time', (4, int)),
                ('modification_time', (4, int)),
                ('timescale', (4, int)),
                ('duration', (4, int))
            ])

        definition.extend([
            ('rate', (4, int)),
            ('volume', (2, int)),
            ('reserved', (2, int)),
            ('preferred_long', (list, (4, int), 2)),
            ('matrix', (list, (4, int), 9)),
            ('preview_time', (4, int)),
            ('preview_duration', (4, int)),
            ('poster_time', (4, int)),
            ('selection_time', (4, int)),
            ('selection_duration', (4, int)),
            ('current_time', (4, int)),
            ('next_track_ID', (4, int))
        ])

        self.properties.update(interpret_atom(atom_body, definition))
        self._definition['MvhdAtom'] = definition

    def to_bytes(self):
        written = FullAtom.to_bytes(self)
        return ''.join([written, write_atom(self.properties, self._definition['MvhdAtom'])])

ATOM_TYPE_TO_CLASS = {
    'free': FreeAtom,
    'ftyp': FtypAtom,
    'mdat': MdatAtom,
    'mvhd': MvhdAtom,
    'skip': SkipAtom,
    'uuid': UserExtendedAtom,
}



import os
import StringIO

from isomedia import atom, isom_atoms
from isomedia.atom import AtomHeader, GenericAtom, ContainerAtom, interpret_int32, interpret_int64
from isomedia.exceptions import MalformedIsomFile, AtomSpecificationError

def get_ptr_size(ptr):
    ptr.seek(0, os.SEEK_END)
    size = ptr.tell()
    ptr.seek(0, os.SEEK_SET)
    return size

def need_read(ptr, n):
    data = ptr.read(n)
    if data is None or len(data) != n:
        raise EOFError
    return data

def interpret_atom_header(data):
    atom_size = interpret_int32(data, 0)
    atom_type = data[4:8]

    if atom_size == 1:
        atom_size = interpret_int64(data, 8)
        header_length = 16
    else:
        header_length = 8

    return (atom_type, atom_size, header_length)

def parse_file(ptr, document):
    atoms = []

    current_offset = 0
    filesize = get_ptr_size(ptr)

    while current_offset < filesize:
        new_atom, atom_size = parse_atom(ptr, current_offset, document=document, parent=None)
        atoms.append(new_atom)
        current_offset += atom_size

    return atoms

def parse_atom(ptr, offset, document=None, parent=None):
    def parse_atom_header(ptr):
        data = need_read(ptr, 8)
        atom_size = interpret_int32(data, 0)
        if atom_size == 1:
            data += need_read(ptr, 8)

        return data

    try:
        atom_header = parse_atom_header(ptr)
    except EOFError:
        raise MalformedIsomFile

    atom_type, atom_size, header_length = interpret_atom_header(atom_header)

    atom_header = AtomHeader(atom_type, atom_size, header_length)
    atom_body_length = atom_size - header_length

    # TODO: Clearly distinguish different atom specifications
    new_atom = None
    atom_body = None

    if atom_type in atom.CONTAINER_ATOMS:
        new_atom = ContainerAtom(atom_header, None, document, parent, offset)
        new_atom.children = parse_children(ptr, offset + header_length, atom_body_length, parent=new_atom)
    elif atom_type in isom_atoms.ATOM_TYPE_TO_CLASS:
        new_atom_class = isom_atoms.ATOM_TYPE_TO_CLASS[atom_type]
        if new_atom_class.LOAD_DATA:
            try:
                atom_body = need_read(ptr, atom_body_length)
            except EOFError:
                raise MalformedIsomFile
        else:
            atom_body = None
            ptr.seek(atom_body_length, os.SEEK_CUR)

        # If the atom is the right size but doesn't match the definition, we can still parse the rest of the file and
        # just default this atom to a GenericAtom and let the caller munge the bits.
        try:
            new_atom = new_atom_class(atom_header, StringIO.StringIO(atom_body), document, parent, offset)
        except AtomSpecificationError:
            new_atom = None

    if new_atom is None:
        try:
            atom_body = atom_body or need_read(ptr, atom_body_length)
        except EOFError:
            raise MalformedIsomFile
        new_atom = GenericAtom(atom_header, StringIO.StringIO(atom_body), document, parent, offset)

    return (new_atom, atom_size)

def parse_children(ptr, offset, total_bytes, parent=None):
    children = []
    bytes_read = 0

    while bytes_read < total_bytes:
        new_atom, atom_size = parse_atom(ptr, offset + bytes_read, parent=parent)
        children.append(new_atom)
        bytes_read += atom_size

    if bytes_read != total_bytes:
        raise MalformedIsomFile

    return children



from itertools import chain
import StringIO
import struct

from isomedia.exceptions import AtomSpecificationError

CONTAINER_ATOMS = [
    '\xa9alb',
    '\xa9art',
    '\xa9cmt',
    '\xa9day',
    '\xa9gen',
    '\xa9grp',
    '\xa9lyr',
    '\xa9nam',
    '\xa9too',
    '\xa9wrt',
    'catg',
    'covr',
    'cpil',
    'cptr',
    'desc',
    'disk',
    'edts',
    'egid',
    'extr',
    'fiin',
    'gnre',
    'ilst',
    'ipro',
    'keyw',
    'mdia',
    'meco',
    'mfra',
    'minf',
    'moof',
    'moov',
    'mvex',
    'paen',
    'pcst',
    'pgap',
    'purd',
    'purl',
    'rinf',
    'rtng',
    'schi',
    'sinf',
    'srpp',
    'stik',
    'strd',
    'strk',
    'stbl',
    'tmpo',
    'traf',
    'trak',
    'tref',
    'trkn',
    'tven',
    'tves',
    'tvnn',
    'tvsh',
    'tvsn',
    'udta',
]

UINT_BYTES_TO_FORMAT = {
    1: '>B',
    2: '>H',
    4: '>I',
    8: '>Q'
}

MAX_UINT32 = (2 ** 32) - 1
STANDARD_HEADER_LENGTH = 8
EXTENDED_HEADER_LENGTH = 16

def interpret_int(data, offset, size):
    assert size in UINT_BYTES_TO_FORMAT
    field = data[offset:offset + size]
    return struct.unpack(UINT_BYTES_TO_FORMAT[size], field)[0]

def interpret_int64(data, offset=0):
    return interpret_int(data, offset, 8)

def interpret_int32(data, offset=0):
    return interpret_int(data, offset, 4)

def interpret_int16(data, offset=0):
    return interpret_int(data, offset, 2)

def interpret_int8(data, offset=0):
    return interpret_int(data, offset, 1)

def write_atom_header(atom):
    header_bytes = ''

    if atom.size > MAX_UINT32:
        header_bytes += struct.pack(UINT_BYTES_TO_FORMAT[4], 1)
    else:
        header_bytes += struct.pack(UINT_BYTES_TO_FORMAT[4], atom.size)
    header_bytes += atom.type
    if atom.size > MAX_UINT32:
        header_bytes += struct.pack(UINT_BYTES_TO_FORMAT[8], atom.size)

    return header_bytes

def interpret_atom(atom_body, definition):
    def cast_field(data, size, cast):
        if cast == int:
            data = interpret_int(data, 0, size)
        return data

    result = {}

    for field in definition:
        field_name, field_type = field
        length = field_type[0]

        if length == list:
            item_length, item_cast = field_type[1]
            item_count = field_type[2]

            data = []
            while item_count is None or len(data) < item_count:
                item_data = atom_body.read(item_length)
                if not item_data:
                    break
                elif len(item_data) != item_length:
                    raise AtomSpecificationError

                item_data = cast_field(item_data, item_length, item_cast)
                data.append(item_data)

        else:
            data = atom_body.read(length)
            if len(data) != length:
                raise AtomSpecificationError

            cast = field_type[1]
            data = cast_field(data, length, cast)

        result[field_name] = data

    return result

def write_atom(properties, definition):
    result = ''

    for field in definition:
        field_name, field_type = field
        field_value = properties[field_name]

        length = field_type[0]

        if length == list:
            item_length, item_cast = field_type[1]

            for item_value in field_value:
                if item_cast == int:
                    result += struct.pack(UINT_BYTES_TO_FORMAT[item_length], item_value)
                else:
                    result += item_value
        else:
            cast = field_type[1]
            if cast == int:
                result += struct.pack(UINT_BYTES_TO_FORMAT[length], field_value)
            else:
                result += field_value

    return result

class AtomHeader(object):
    def __init__(self, atom_type, atom_size, header_length):
        self.type = atom_type
        self.size = atom_size
        self.header_length = header_length

class Atom(object):
    # Allow lazy-loading (enabled when False)
    LOAD_DATA = True

    def __init__(self, atom_header, atom_body, document, parent_atom, file_offset):
        self.header = atom_header

        self.document = document
        self.parent_atom = parent_atom

        self._input_file_offset = file_offset
        self._input_size = atom_header.size
        self._body_offset = atom_header.header_length

        self._definition = {}
        self.properties = {}

    def to_bytes(self):
        return write_atom_header(self)

    def __repr__(self):
        return str({
            'type': self.type
        })

    @property
    def type(self):
        return self.header.type

    @type.setter
    def type(self, value):
        assert isinstance(value, str), 'A box\'s type field should be a string.'
        assert len(value) == 4, 'A box\'s type should be 4 bytes exactly.'
        self.header.type = value

    @property
    def size(self):
        return self.header.size

    @size.setter
    def size(self, value):
        self.header.size = value

class FullAtom(Atom):
    def __init__(self, atom_header, atom_body, document, parent_atom, file_offset):
        Atom.__init__(self, atom_header, atom_body, document, parent_atom, file_offset)

        definition = [
            ('version', (1, int)),
            ('flags', (3, None))
        ]

        self.properties.update(interpret_atom(atom_body, definition))
        self._definition['FullAtom'] = definition

    def to_bytes(self):
        written = Atom.to_bytes(self)
        return ''.join([written, write_atom(self.properties, self._definition['FullAtom'])])

class ContainerAtom(Atom):
    def __init__(self, atom_header, atom_body, document, parent_atom, file_offset):
        Atom.__init__(self, atom_header, atom_body, document, parent_atom, file_offset)
        self.children = []

    def __repr__(self):
        return str({
            'type': self.type,
            'children': self.children
        })

    def to_bytes(self):
        written = Atom.to_bytes(self)
        return ''.join(chain([written], (child.to_bytes() for child in self.children)))

    def append_child(self, child):
        self.size += child.size
        self.children.append(child)

class GenericAtom(Atom):
    def __init__(self, atom_header, atom_body, document, parent_atom, file_offset):
        Atom.__init__(self, atom_header, atom_body, document, parent_atom, file_offset)
        self._data = atom_body.read()

    def get_data(self):
        return self._data

    def to_bytes(self):
        written = Atom.to_bytes(self)
        return ''.join([written, self.get_data()])

class LazyLoadAtom(Atom):
    LOAD_DATA = False

    def __init__(self, atom_header, atom_body, document, parent_atom, file_offset):
        Atom.__init__(self, atom_header, atom_body, document, parent_atom, file_offset)
        self._data = None

    def get_data(self):
        if self._data is None:
            fp = self.document.fp
            fp.seek(self._input_file_offset + self._body_offset)
            self._data = fp.read(self._input_size - self._body_offset)

        return self._data

    def to_bytes(self):
        written = Atom.to_bytes(self)
        return ''.join([written, self.get_data()])

def create_atom(atom_type, atom_body):
    body_length = len(atom_body)
    if body_length + STANDARD_HEADER_LENGTH <= MAX_UINT32:
        header_length = STANDARD_HEADER_LENGTH
    else:
        header_length = EXTENDED_HEADER_LENGTH

    atom_size = header_length + body_length
    atom_header = AtomHeader(None, atom_size, header_length)
    atom = GenericAtom(atom_header, StringIO.StringIO(atom_body), None, None, None)
    atom.type = atom_type

    return atom



from isomedia.atom import ContainerAtom, write_atom_header
from isomedia.exceptions import MalformedIsomFile
from isomedia.parser import parse_file

CHUNK_SIZE = 1024

class ISOBaseMediaFile(object):
    def __init__(self, fp):
        self.fp = fp
        self.atoms = parse_file(fp, self)

    def __write_atom(self, atom, fp):
        if isinstance(atom, ContainerAtom):
            fp.write(write_atom_header(atom))

            for child in atom.children:
                self.__write_atom(child, fp)
        else:
            if atom.LOAD_DATA:
                fp.write(atom.to_bytes())
            else:
                # Lazy-loaded Atoms are read-and-copy only.
                self.fp.seek(atom._input_file_offset)
                remaining_to_read = atom.size

                while remaining_to_read > 0:
                    next_chunk_length = min(CHUNK_SIZE, remaining_to_read)
                    fp.write(self.fp.read(next_chunk_length))
                    remaining_to_read -= next_chunk_length

    def write(self, fp):
        for atom in self.atoms:
            self.__write_atom(atom, fp)

    def __repr__(self):
        return str(self.atoms)

def load(fp):
    return ISOBaseMediaFile(fp)



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='isomedia',
    version='0.2.1',
    description='This is another ISO base media format file parser.',
    author='Felix Fung',
    author_email='felix.the.cheshire.cat@gmail.com',
    url='https://github.com/flxf/isomedia',
    packages=find_packages(exclude=['tests']),
    install_requires=[],

    keywords=['mp4', 'isom'],
)
